#=======================================================================
# simple_reg_test.py
#=======================================================================

from __future__ import absolute_import, division, print_function

import os

import pytest

from pymtl3 import *
from pymtl3.passes.PassGroups import *
from pymtl3.passes.sverilog.import_.ImportPass import ImportPass


def get_dir():
  return os.path.dirname(os.path.abspath(__file__))+os.path.sep

@pytest.mark.xfail( reason = "port mapping not supported yet" )
def test_VReg( dump_vcd ):

  class VReg( Component ):

    def construct( s ):
      s.in_ = InPort ( Bits32 )
      s.out = OutPort( Bits32 )

      s.sv_port_mapping = {
        'clk' : s.clk,
        'd'   : s.in_,
        'q'   : s.out,
      }

  class Wrapper( Component ):
    def construct( s ):
      s.in_ = InPort ( Bits32 )
      s.out = OutPort( Bits32 )

      s.inner = VReg()( in_ = s.in_, out = s.out )

    def line_trace( s ):
      return "{} > {}".format( s.in_, s.out )

  #---------------------------------------------------------------------
  # test
  #---------------------------------------------------------------------

  m = Wrapper()
  m.sverilog_import_path = get_dir()+'VReg.sv'
  dtype = Bits32

  m.elaborate()

  m.apply( ImportPass() ),

  m.apply( GenDAGPass() ),
  m.apply( WrapGreenletPass() ),
  m.apply( DynamicSchedulePass() ),
  m.apply( SimpleTickPass() ),
  m.lock_in_simulation()

  m.sim_reset()

  for i in range( 10 ):
    m.in_ = dtype( i )
    m.tick()
    print(m.line_trace())
    assert m.out == i



#=========================================================================
# StructuralTranslatorL1.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 24, 2019
"""Provide L1 structural translator."""

from __future__ import absolute_import, division, print_function

from collections import defaultdict, deque
from functools import reduce

from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.rtlir import StructuralRTLIRSignalExpr as sexp
from pymtl3.passes.rtlir.structural.StructuralRTLIRGenL1Pass import (
    StructuralRTLIRGenL1Pass,
)

from ..BaseRTLIRTranslator import BaseRTLIRTranslator, TranslatorMetadata


def gen_connections( top ):
  """Return a tuple of all connections in the hierarchy whose top is `top`.
  
  Return a three element tuple ( ss, sc, cc ). `ss` is a dictionary indexed
  by component `m` and has a set of pairs of connected signals within component
  `m` ( and thus is called "self_self" ). `sc` is a dictionary indexed by
  component `m` and has a set of pairs of connected signals between component
  `m` and its child components ( "self_child" ). `cc` is a dictionary indexed
  by component `m` and has a set of pairs of connected signals between two
  child components of `m` ( "child_child" ).
  """
  _top_conns_self_self = defaultdict( set )
  _top_conns_self_child = defaultdict( set )
  _top_conns_child_child = defaultdict( set )

  nets = top.get_all_value_nets()
  adjs = top.get_signal_adjacency_dict()

  for writer, net in nets:
    S = deque( [ writer ] )
    visited = {  writer  }
    while S:
      u = S.pop()
      writer_host        = u.get_host_component()
      writer_host_parent = writer_host.get_parent_object()
      for v in adjs[u]:
        if v not in visited:
          visited.add( v )
          S.append( v )
          reader_host        = v.get_host_component()
          reader_host_parent = reader_host.get_parent_object()

          # Four possible cases for the reader and writer signals:
          # 1.   They have the same host component. Both need
          #       to be added to the host component.
          # 2/3. One's host component is the parent of the other.
          #       Both need to be added to the parent component.
          # 4.   They have the same parent component.
          #       Both need to be added to the parent component.

          if writer_host is reader_host:
            _top_conns_self_self[writer_host].add( ( u, v ) )
          elif writer_host_parent is reader_host:
            _top_conns_self_child[reader_host].add( ( u, v ) )
          elif writer_host is reader_host_parent:
            _top_conns_self_child[writer_host].add( ( u, v ) )
          elif writer_host_parent == reader_host_parent:
            _top_conns_child_child[writer_host_parent].add( ( u, v ) )
          else:
            assert False, "unexpected connection type!"

  return \
    _top_conns_self_self, _top_conns_self_child, _top_conns_child_child

class StructuralTranslatorL1( BaseRTLIRTranslator ):
  def __init__( s, top ):
    super( StructuralTranslatorL1, s ).__init__( top )
    # To avoid doing redundant computation, we generate the connections of
    # the entire hierarchy once and only once here.
    # c_ss: self-self connections
    # c_sc: self-child connections
    # c_cc: child-child connections
    s.c_ss, s.c_sc, s.c_cc = gen_connections( top )

  def clear( s, tr_top ):
    super( StructuralTranslatorL1, s ).clear( tr_top )
    # Metadata namespace for RTLIR structural translator and the backend
    # structural translator
    s.structural = TranslatorMetadata()
    s.s_backend = TranslatorMetadata()

    # Generate metadata
    s.gen_structural_trans_metadata( tr_top )

    # Data type declaration
    s.structural.decl_type_vector = []
    s.structural.decl_type_array  = []

  #-----------------------------------------------------------------------
  # gen_structural_trans_metadata
  #-----------------------------------------------------------------------

  def gen_structural_trans_metadata( s, tr_top ):
    # c_ss: self-self connections
    # c_sc: self-child connections
    # c_cc: child-child connections
    tr_top.apply( StructuralRTLIRGenL1Pass( s.c_ss, s.c_sc, s.c_cc ) )

  #-----------------------------------------------------------------------
  # translate_structural
  #-----------------------------------------------------------------------

  def translate_structural( s, tr_top ):
    """Translate structural part of top component under translation.

    This function will only be called once during the whole translation
    process.
    """
    # Component metadata
    s.structural.component_name = {}
    s.structural.component_file_info = {}
    s.structural.component_unique_name = {}

    # Declarations
    s.structural.decl_ports  = {}
    s.structural.decl_wires  = {}
    s.structural.decl_consts = {}

    # Connections
    s.structural.connections = {}
    s._translate_structural( tr_top )

  #-----------------------------------------------------------------------
  # _translate_structural
  #-----------------------------------------------------------------------

  def _translate_structural( s, m ):
    """Translate structural part of component m.

    This method will be recursively applied to different components in the
    hierarchy.
    """
    m_rtype = m._pass_structural_rtlir_gen.rtlir_type
    s.structural.component_file_info[m] = m_rtype.get_file_info()
    s.structural.component_name[m] = m_rtype.get_name()
    s.structural.component_unique_name[m] = \
        s.rtlir_tr_component_unique_name(m_rtype)

    # Translate declarations of signals
    s.translate_decls( m )

    # Translate connections
    s.translate_connections( m )

  #-----------------------------------------------------------------------
  # translate_decls
  #-----------------------------------------------------------------------

  def translate_decls( s, m ):
    m_rtype = m._pass_structural_rtlir_gen.rtlir_type

    # Ports
    port_decls = []
    for port_id, rtype in m_rtype.get_ports_packed():
      if isinstance( rtype, rt.Array ):
        array_type = rtype
        port_rtype = rtype.get_sub_type()
      else:
        array_type = None
        port_rtype = rtype
      port_decls.append(
        s.rtlir_tr_port_decl(
          s.rtlir_tr_var_id( port_id ),
          port_rtype,
          s.rtlir_tr_unpacked_array_type( array_type ),
          s.rtlir_data_type_translation( m, port_rtype.get_dtype() )
      ) )
    s.structural.decl_ports[m] = s.rtlir_tr_port_decls( port_decls )

    # Wires
    wire_decls = []
    for wire_id, rtype in m_rtype.get_wires_packed():
      if isinstance( rtype, rt.Array ):
        array_type = rtype
        wire_rtype = rtype.get_sub_type()
      else:
        array_type = None
        wire_rtype = rtype
      wire_decls.append(
        s.rtlir_tr_wire_decl(
          s.rtlir_tr_var_id( wire_id ),
          wire_rtype,
          s.rtlir_tr_unpacked_array_type( array_type ),
          s.rtlir_data_type_translation( m, wire_rtype.get_dtype() )
      ) )
    s.structural.decl_wires[m] = s.rtlir_tr_wire_decls( wire_decls )

    # Consts
    const_decls = []
    if hasattr( s, "behavioral" ):
      used_set = s.behavioral.accessed[m]
    else:
      used_set = None

    for const_id, rtype, instance in m._pass_structural_rtlir_gen.consts:
      if used_set is None or const_id in used_set:
        if isinstance( rtype, rt.Array ):
          array_type = rtype
          const_rtype = rtype.get_sub_type()
        else:
          array_type = None
          const_rtype = rtype
        const_decls.append(
          s.rtlir_tr_const_decl(
            s.rtlir_tr_var_id( const_id ),
            const_rtype,
            s.rtlir_tr_unpacked_array_type( array_type ),
            s.rtlir_data_type_translation( m, const_rtype.get_dtype() ),
            instance
        ) )
    s.structural.decl_consts[m] = s.rtlir_tr_const_decls( const_decls )

  #-----------------------------------------------------------------------
  # translate_connections
  #-----------------------------------------------------------------------

  def translate_connections( s, m ):
    connections = []
    _connections = m._pass_structural_rtlir_gen.connections
    for writer, reader in _connections:
      connections.append( s.rtlir_tr_connection(
        s.rtlir_signal_expr_translation( writer, m ),
        s.rtlir_signal_expr_translation( reader, m )
      ) )
    s.structural.connections[m] = s.rtlir_tr_connections( connections )

  #-----------------------------------------------------------------------
  # rtlir_data_type_translation
  #-----------------------------------------------------------------------

  def rtlir_data_type_translation( s, m, dtype ):
    """Translate an RTLIR data type into its backend representation."""
    if isinstance( dtype, rdt.Vector ):
      ret = s.rtlir_tr_vector_dtype( dtype )
      if reduce( lambda r, x: r and dtype != x[0],
          s.structural.decl_type_vector, True ):
        s.structural.decl_type_vector.append( ( dtype, ret ) )
      return ret

    else:
      assert False, "unsupported RTLIR dtype {} at L1!".format( dtype )

  #-----------------------------------------------------------------------
  # rtlir_signal_expr_translation
  #-----------------------------------------------------------------------

  def rtlir_signal_expr_translation( s, expr, m ):
    """Translate a signal expression in RTLIR into its backend representation.

    Only the following operations are supported at L1:
    sexp.CurComp, sexp.CurCompAttr, sexp.BitSelection, sexp.PartSelection, sexp.PortIndex,
    sexp.WireIndex, sexp.ConstIndex
    """
    if isinstance( expr, sexp.CurComp ):
      comp_id, comp_rtype = expr.get_component_id(), expr.get_rtype()
      return s.rtlir_tr_current_comp( comp_id, comp_rtype )

    elif isinstance( expr, sexp.CurCompAttr ):
      return s.rtlir_tr_current_comp_attr(
        s.rtlir_signal_expr_translation( expr.get_base(), m ),
        expr.get_attr() )

    elif isinstance( expr, sexp.PortIndex ):
      return s.rtlir_tr_port_array_index(
        s.rtlir_signal_expr_translation( expr.get_base(), m ),
        expr.get_index() )

    elif isinstance( expr, sexp.WireIndex ):
      return s.rtlir_tr_wire_array_index(
        s.rtlir_signal_expr_translation( expr.get_base(), m ),
        expr.get_index() )

    elif isinstance( expr, sexp.ConstIndex ):
      return s.rtlir_tr_const_array_index(
        s.rtlir_signal_expr_translation( expr.get_base(), m ),
        expr.get_index() )

    elif isinstance( expr, sexp.BitSelection ):
      base = expr.get_base()
      assert not isinstance(base, (sexp.PartSelection, sexp.BitSelection)), \
        'bit selection {} over bit/part selection {} is not allowed!'.format(expr, base)
      return s.rtlir_tr_bit_selection(
        s.rtlir_signal_expr_translation( expr.get_base(), m ), expr.get_index() )

    elif isinstance( expr, sexp.PartSelection ):
      base = expr.get_base()
      assert not isinstance(base, (sexp.PartSelection, sexp.BitSelection)), \
        'part selection {} over bit/part selection {} is not allowed!'.format(expr, base)
      start, stop = expr.get_slice()[0], expr.get_slice()[1]
      return s.rtlir_tr_part_selection(
        s.rtlir_signal_expr_translation( expr.get_base(), m ), start, stop )

    elif isinstance( expr, sexp.ConstInstance ):
      dtype = expr.get_rtype().get_dtype()
      assert isinstance( dtype, rdt.Vector ), \
          '{} is not supported at L1!'.format( dtype )
      return s.rtlir_tr_literal_number( dtype.get_length(), expr.get_value() )

    # Other operations are not supported at L1
    else:
      assert False, '{} is not supported at L1!'.format( expr )

  #-----------------------------------------------------------------------
  # Methods to be implemented by the backend translator
  #-----------------------------------------------------------------------

  # Data types
  def rtlir_tr_vector_dtype( s, Type ):
    raise NotImplementedError()

  def rtlir_tr_unpacked_array_type( s, Type ):
    raise NotImplementedError()

  # Declarations
  def rtlir_tr_port_decls( s, port_decls ):
    raise NotImplementedError()

  def rtlir_tr_port_decl( s, id_, rtype, array_type, dtype ):
    raise NotImplementedError()

  def rtlir_tr_wire_decls( s, wire_decls ):
    raise NotImplementedError()

  def rtlir_tr_wire_decl( s, id_, Type, array_type, dtype ):
    raise NotImplementedError()

  def rtlir_tr_const_decls( s, const_decls ):
    raise NotImplementedError()

  def rtlir_tr_const_decl( s, id_, Type, array_type, dtype, value ):
    raise NotImplementedError()

  # Connections
  def rtlir_tr_connections( s, connections ):
    raise NotImplementedError()

  def rtlir_tr_connection( s, wr_signal, rd_signal ):
    raise NotImplementedError()

  # Signal operations
  def rtlir_tr_bit_selection( s, base_signal, index ):
    raise NotImplementedError()

  def rtlir_tr_part_selection( s, base_signal, start, stop ):
    raise NotImplementedError()

  def rtlir_tr_port_array_index( s, base_signal, index ):
    raise NotImplementedError()

  def rtlir_tr_wire_array_index( s, base_signal, index ):
    raise NotImplementedError()

  def rtlir_tr_const_array_index( s, base_signal, index ):
    raise NotImplementedError()

  def rtlir_tr_current_comp_attr( s, base_signal, attr ):
    raise NotImplementedError()

  def rtlir_tr_current_comp( s, comp_id, comp_rtype ):
    raise NotImplementedError()

  # Miscs
  def rtlir_tr_var_id( s, var_id ):
    raise NotImplementedError()

  def rtlir_tr_literal_number( s, nbits, value ):
    raise NotImplementedError()

  def rtlir_tr_component_unique_name( s, c_rtype ):
    raise NotImplementedError()



#=========================================================================
# ImportedObject_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 2, 2019
"""Test if the imported object works correctly."""

from __future__ import absolute_import, division, print_function

import os

from pymtl3.datatypes import Bits1, Bits32, Bits64
from pymtl3.dsl import Component, InPort, Interface, OutPort
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.passes.sverilog.import_.ImportPass import ImportPass
from pymtl3.stdlib.test import TestVectorSimulator


def get_dir():
  return os.path.dirname(os.path.abspath(__file__))+os.path.sep

def local_do_test( _m ):
  _m.elaborate()
  ipass = ImportPass()
  m = ipass.get_imported_object( _m )
  sim = TestVectorSimulator( m, _m._test_vectors, _m._tv_in, _m._tv_out )
  sim.run_test()

def test_reg( do_test ):
  def tv_in( m, test_vector ):
    m.d = Bits32( test_vector[0] )
  def tv_out( m, test_vector ):
    if test_vector[1] != '*':
      assert m.q == Bits32( test_vector[1] )
  class VReg( Component ):
    def construct( s ):
      s.q = OutPort( Bits32 )
      s.d = InPort( Bits32 )
  a = VReg()
  a.sverilog_import_path = get_dir()+'VReg.sv'
  a._test_vectors = [
    [    1,    '*' ],
    [    2,      1 ],
    [   -1,      2 ],
    [   -2,     -1 ],
    [   42,     -2 ],
    [  -42,     42 ],
  ]
  a._tv_in = tv_in
  a._tv_out = tv_out
  do_test( a )

def test_adder( do_test ):
  def tv_in( m, test_vector ):
    m.in0 = Bits32( test_vector[0] )
    m.in1 = Bits32( test_vector[1] )
    m.cin = Bits1( test_vector[2] )
  def tv_out( m, test_vector ):
    if test_vector[3] != '*':
      assert m.out == Bits32( test_vector[3] )
    if test_vector[4] != '*':
      assert m.cout == Bits32( test_vector[4] )
  class VAdder( Component ):
    def construct( s ):
      s.in0 = InPort( Bits32 )
      s.in1 = InPort( Bits32 )
      s.cin = InPort( Bits1 )
      s.out = OutPort( Bits32 )
      s.cout = OutPort( Bits1 )
  a = VAdder()
  a.sverilog_import_path = get_dir()+'VAdder.sv'
  a._test_vectors = [
    [    1,      1,     1,     3, 0 ],
    [    1,     -1,     0,     0, 1 ],
    [   42,     42,     1,    85, 0 ],
    [   42,    -43,     1,     0, 1 ],
  ]
  a._tv_in = tv_in
  a._tv_out = tv_out
  do_test( a )



#=========================================================================
# SVBehavioralTranslatorL1.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 18, 2019
"""Provide the level 1 SystemVerilog translator implementation."""
from __future__ import absolute_import, division, print_function

import inspect

from pymtl3.datatypes import Bits, Bits32
from pymtl3.passes.rtlir import BehavioralRTLIR as bir
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.sverilog.errors import SVerilogTranslationError
from pymtl3.passes.sverilog.util.utility import make_indent
from pymtl3.passes.translator.behavioral.BehavioralTranslatorL1 import (
    BehavioralTranslatorL1,
)

from .SVBehavioralTranslatorL0 import SVBehavioralTranslatorL0


class SVBehavioralTranslatorL1( SVBehavioralTranslatorL0, BehavioralTranslatorL1 ):

  def rtlir_tr_upblk_decls( s, upblk_decls ):
    ret = ''
    for idx, upblk_decl in enumerate(upblk_decls):
      make_indent( upblk_decl, 1 )
      if idx != 0:
        ret += '\n'
      ret += '\n' + '\n'.join( upblk_decl )
    return ret

  def rtlir_tr_upblk_decl( s, upblk, src, py_src ):
    return py_src + [ "" ] + src

  def _get_rtlir2sv_visitor( s ):
    return BehavioralRTLIRToSVVisitorL1

  def rtlir_tr_upblk_srcs( s, upblk_srcs ):
    ret = ''
    for upblk_src in upblk_srcs:
      make_indent( upblk_src, 1 )
      ret += '\n' + '\n'.join( upblk_src )
    return ret

  def rtlir_tr_upblk_src( s, upblk, rtlir_upblk ):
    visitor = s._get_rtlir2sv_visitor()(s.is_sverilog_reserved)
    return visitor.enter( upblk, rtlir_upblk )

  def rtlir_tr_upblk_py_srcs( s, upblk_py_srcs ):
    ret = ''
    for upblk_py_src in upblk_py_srcs:
      make_indent( upblk_py_src, 1 )
      ret += '\n' + '\n'.join( upblk_py_src )
    return ret

  def rtlir_tr_upblk_py_src( s, upblk ):
    def _trim( py_src ):
      indent = 100
      for line in py_src:
        if line:
          n_spaces = len( line ) - len( line.lstrip() )
          if n_spaces < indent:
            indent = n_spaces
      for idx, line in enumerate( py_src ):
        if line:
          py_src[ idx ] = line[indent:]
    try:
      upblk_py_src = inspect.getsource( upblk )
    except IOError:
      upblk_py_src = "IOError: cannot retrieve Python update block code!"
    upblk_py_src = upblk_py_src.split( '\n' )
    _trim( upblk_py_src )
    while upblk_py_src and not upblk_py_src[-1]:
      upblk_py_src = upblk_py_src[:-1]
    py_src = [ "PYMTL SOURCE:", "" ] + upblk_py_src
    return map( lambda x: "// "+x, py_src )

  def rtlir_tr_behavioral_freevars( s, freevars ):
    make_indent( freevars, 1 )
    return '\n'.join( freevars )

  def rtlir_tr_behavioral_freevar( s, id_, rtype, array_type, dtype, obj ):
    assert isinstance( rtype, rt.Const ), \
      '{} freevar should be a constant!'.format( id_ )
    assert isinstance( rtype.get_dtype(), rdt.Vector ), \
      '{} freevar should be a (list of) integer/BitStruct!'.format( id_ )
    return s.rtlir_tr_const_decl( '__const$'+id_, rtype, array_type, dtype, obj )

#-------------------------------------------------------------------------
# BehavioralRTLIRToSVVisitorL1
#-------------------------------------------------------------------------

class BehavioralRTLIRToSVVisitorL1( bir.BehavioralRTLIRNodeVisitor ):
  """Visitor that translates RTLIR to SystemVerilog for a single upblk."""

  def __init__( s, is_reserved ):
    # Should use enum here, but enum is a python 3 feature...
    s.NONE          = 0
    s.COMBINATIONAL = 1
    s.SEQUENTIAL    = 2
    s.upblk_type    = s.NONE
    s._is_sverilog_reserved = is_reserved

  def is_sverilog_reserved( s, name ):
    return s._is_sverilog_reserved( name )

  def enter( s, blk, rtlir ):
    """Entry point for RTLIR generation."""
    s.blk     = blk

    # s.globals contains a dict of the global namespace of the module where
    # blk was defined
    s.globals = blk.__globals__

    # s.closure contains the free variables defined in an enclosing scope.
    # Basically this is the model instance s.
    s.closure = {}

    for i, var in enumerate( blk.__code__.co_freevars ):
      try: 
        s.closure[ var ] = blk.__closure__[ i ].cell_contents
      except ValueError: 
        pass

    return s.visit( rtlir )

  def check_res( s, node, name ):
    if s.is_sverilog_reserved( name ):
      raise SVerilogTranslationError( s.blk, node, 
        "name {} is a SystemVerilog reserved keyword!".format( name ) )

  #-----------------------------------------------------------------------
  # visit_CombUpblk
  #-----------------------------------------------------------------------

  def visit_CombUpblk( s, node ):
    """Return the SV representation of statements inside it."""
    blk_name = node.name
    src      = []
    body     = []
    s.upblk_type = s.COMBINATIONAL

    s.check_res( node, blk_name )

    # Add name of the upblk to this always block
    src.extend( [ 'always_comb begin : {blk_name}'.format( **locals() ) ] )
    
    for stmt in node.body:
      body.extend( s.visit( stmt ) )

    make_indent( body, 1 )
    src.extend( body )
    src.extend( [ 'end' ] )
    s.upblk_type = s.NONE
    return src

  #-----------------------------------------------------------------------
  # visit_SeqUpblk
  #-----------------------------------------------------------------------

  def visit_SeqUpblk( s, node ):
    """Return the SV representation of statements inside it."""
    blk_name = node.name
    src      = []
    body     = []
    s.upblk_type = s.SEQUENTIAL

    s.check_res( node, blk_name )

    # Add name of the upblk to this always block
    src.extend( [
      'always_ff @(posedge clk) begin : {blk_name}'.format( **locals() )
    ] )
    
    for stmt in node.body:
      body.extend( s.visit( stmt ) )

    make_indent( body, 1 )
    src.extend( body )
    src.extend( [ 'end' ] )
    s.upblk_type = s.NONE
    return src

  #-----------------------------------------------------------------------
  # Statements
  #-----------------------------------------------------------------------
  # All statement nodes return a list of strings.

  #-----------------------------------------------------------------------
  # visit_Assign
  #-----------------------------------------------------------------------

  def visit_Assign( s, node ):
    target        = s.visit( node.target )
    value         = s.visit( node.value )
    assignment_op = '<=' if s.upblk_type == s.SEQUENTIAL else '='
    ret = '{target} {assignment_op} {value};'.format( **locals() )
    return [ ret ]

  #-----------------------------------------------------------------------
  # visit_If
  #-----------------------------------------------------------------------

  def visit_If( s, node ):
    raise SVerilogTranslationError( s.blk, node, "If not supported at L1" )

  #-----------------------------------------------------------------------
  # visit_For
  #-----------------------------------------------------------------------

  def visit_For( s, node ):
    raise SVerilogTranslationError( s.blk, node, "For not supported at L1" )

  #-----------------------------------------------------------------------
  # Expressions
  #-----------------------------------------------------------------------
  # All expression nodes return a single string.  

  #-----------------------------------------------------------------------
  # visit_Number
  #-----------------------------------------------------------------------

  def visit_Number( s, node ):
    """Return a number in string without width specifier."""
    return str( node.value )

  #-----------------------------------------------------------------------
  # visit_Concat
  #-----------------------------------------------------------------------

  def visit_Concat( s, node ):
    values = [ s.visit(v) for v in node.values ]
    signals = ', '.join( values )
    return '{{ {signals} }}'.format( **locals() )

  #-----------------------------------------------------------------------
  # visit_ZeroExt
  #-----------------------------------------------------------------------

  def visit_ZeroExt( s, node ):
    value = s.visit( node.value )
    try:
      target_nbits = int(node.nbits._value)
    except AttributeError:
      raise SVerilogTranslationError( s.blk, node, 
        "new bitwidth of zero extension must be known at elaboration time!" )
    current_nbits = int(node.value.Type.get_dtype().get_length())
    padded_nbits = target_nbits - current_nbits
    return "{{ {{ {padded_nbits} {{ 1'b0 }} }}, {value} }}".format( **locals() )

  #-----------------------------------------------------------------------
  # visit_SignExt
  #-----------------------------------------------------------------------
  # We need to special case the situation where a bit-selection or single-bit
  # part selection needs to be sign-extended ( Verilator throws an error
  # at this ) e.g. in_[31:31][0] should be translated into in_[31:31].

  def visit_SignExt( s, node ):
    value = s.visit( node.value )
    try:
      target_nbits = int(node.nbits._value)
    except AttributeError:
      raise SVerilogTranslationError( s.blk, node, 
        "new bitwidth of sign extension must be known at elaboration time!" )

    current_nbits = int(node.value.Type.get_dtype().get_length())
    last_bit = current_nbits - 1
    padded_nbits = target_nbits - current_nbits

    template = "{{ {{ {padded_nbits} {{ {value}[{last_bit}] }} }}, {value} }}"
    one_bit_template = "{{ {{ {padded_nbits} {{ {_value} }} }}, {value} }}"

    # Check if the signal to be extended is a bit selection or one-bit part
    # selection.
    if isinstance( node.value, bir.Slice ):
      try:
        lower = node.value.lower._value
        upper = node.value.upper._value
        if upper - lower == 1:
          _one_bit = True
        else:
          _one_bit = False
      except AttributeError:
        _one_bit = False

      # Manipulate the slicing string to avoid indexing on a sliced signal
      if not _one_bit:
        l, col, r = value.rfind('['), value.rfind(':'), value.rfind(']')
        if -1 < l < col < r:
          _value = value[:col] + ']'
          return one_bit_template.format( **locals() )

    elif isinstance( node.value, bir.Index ):
      _one_bit = True
    else:
      _one_bit = False

    if _one_bit:
      _value = value
      return one_bit_template.format( **locals() )
    else:
      return template.format( **locals() )

  #-----------------------------------------------------------------------
  # visit_Reduce
  #-----------------------------------------------------------------------

  def visit_Reduce( s, node ):
    op_t = type(node.op)
    reduce_ops = { bir.BitAnd : '&', bir.BitOr : '|', bir.BitXor : '^' }
    if op_t not in reduce_ops:
      raise SVerilogTranslationError( s.blk, node,
          "unrecognized operator {} for reduce method!".format(op_t) )
    value = s.visit( node.value )
    op = reduce_ops[ op_t ]
    return "( {op} {value} )".format( **locals() )

  #-----------------------------------------------------------------------
  # visit_SizeCast
  #-----------------------------------------------------------------------

  def visit_SizeCast( s, node ):
    nbits = node.nbits
    value = s.visit( node.value )
    if hasattr( node, "_value" ):
      template = "{nbits}'d{value}"
      if isinstance( node._value, Bits ):
        value = node._value.value
      else:
        value = node._value
    else:
      template = "{nbits}'( {value} )"
    return template.format( **locals() )

  #-----------------------------------------------------------------------
  # visit_StructInst
  #-----------------------------------------------------------------------

  def visit_StructInst( s, node ):
    raise SVerilogTranslationError( s.blk, node, "StructInst not supported at L1" )

  #-----------------------------------------------------------------------
  # visit_IfExp
  #-----------------------------------------------------------------------

  def visit_IfExp( s, node ):
    raise SVerilogTranslationError( s.blk, node, "IfExp not supported at L1" )

  #-----------------------------------------------------------------------
  # visit_UnaryOp
  #-----------------------------------------------------------------------

  def visit_UnaryOp( s, node ):
    raise SVerilogTranslationError( s.blk, node, "UnaryOp not supported at L1" )

  #-----------------------------------------------------------------------
  # visit_BoolOp
  #-----------------------------------------------------------------------

  def visit_BoolOp( s, node ):
    raise SVerilogTranslationError( s.blk, node, "BoolOp not supported at L1" )

  #-----------------------------------------------------------------------
  # visit_BinOp
  #-----------------------------------------------------------------------

  def visit_BinOp( s, node ):
    raise SVerilogTranslationError( s.blk, node, "BinOp not supported at L1" )

  #-----------------------------------------------------------------------
  # visit_Compare
  #-----------------------------------------------------------------------

  def visit_Compare( s, node ):
    raise SVerilogTranslationError( s.blk, node, "Compare not supported at L1" )

  #-----------------------------------------------------------------------
  # visit_Attribute
  #-----------------------------------------------------------------------

  def visit_Attribute( s, node ):

    attr  = node.attr
    value = s.visit( node.value )

    s.check_res( node, attr )

    if isinstance( node.value, bir.Base ):
      # The base of this attribute node is the component 's'.
      # Example: s.out, s.STATE_IDLE
      # assert node.value.base is s.component
      ret = attr
    else:
      raise SVerilogTranslationError( s.blk, node, 
          "sub-components are not supported at L1" )

    return ret

  #-----------------------------------------------------------------------
  # visit_Index
  #-----------------------------------------------------------------------

  def visit_Index( s, node ):
    idx   = s.visit( node.idx )
    value = s.visit( node.value )
    Type = node.value.Type

    # Unpacked index
    if isinstance( Type, rt.Array ):

      subtype = Type.get_sub_type()
      # Unpacked array index must be a static constant integer!
      # if idx is None and not isinstance(subtype, (rt.Port, rt.Wire, rt.Const)):
        # raise SVerilogTranslationError( s.blk, node.ast,
# 'index of unpacked array {} must be a constant integer expression!'. \
            # format(node.value) )

      if isinstance( subtype, ( rt.Port, rt.Wire, rt.Const ) ):
        return '{value}[{idx}]'.format( **locals() )
      else:
        return '{value}$__{idx}'.format( **locals() )

    # Index on a signal
    elif isinstance( Type, rt.Signal ):

      # Packed index
      if Type.is_packed_indexable():
        return '{value}[{idx}]'.format( **locals() )
      # Bit selection
      elif isinstance( Type.get_dtype(), rdt.Vector ):
        return '{value}[{idx}]'.format( **locals() )
      else:
        raise SVerilogTranslationError( s.blk, node, 
            "internal error: unrecognized index" )

    else:
      raise SVerilogTranslationError( s.blk, node, 
          "internal error: unrecognized index" )

  #-----------------------------------------------------------------------
  # visit_Slice
  #-----------------------------------------------------------------------

  def visit_Slice( s, node ):

    lower = s.visit( node.lower )
    value = s.visit( node.value )

    if hasattr( node.upper, '_value' ):
      upper = str( int( node.upper._value - Bits32(1) ) )
    else:
      upper = s.visit( node.upper ) + '-1'

    return '{value}[{upper}:{lower}]'.format( **locals() )

  #-----------------------------------------------------------------------
  # visit_Base
  #-----------------------------------------------------------------------

  def visit_Base( s, node ):
    return str( node.base )

  #-----------------------------------------------------------------------
  # visit_LoopVar
  #-----------------------------------------------------------------------

  def visit_LoopVar( s, node ):
    raise SVerilogTranslationError( s.blk, node, "LoopVar not supported at L1" )

  #-----------------------------------------------------------------------
  # visit_FreeVar
  #-----------------------------------------------------------------------

  def visit_FreeVar( s, node ):
    return '__const$'+node.name

  #-----------------------------------------------------------------------
  # visit_TmpVar
  #-----------------------------------------------------------------------

  def visit_TmpVar( s, node ):
    raise SVerilogTranslationError( s.blk, node, "TmpVar not supported at L1" )

  #-----------------------------------------------------------------------
  # visit_LoopVarDecl
  #-----------------------------------------------------------------------

  def visit_LoopVarDecl( s, node ):
    raise SVerilogTranslationError( s.blk, node, "LoopVarDecl not supported at L1" )



#=========================================================================
# StructuralRTLIRGenL1Pass.py
#=========================================================================
# Author : Shunning Jiang, Peitian Pan
# Date   : Apr 3, 2019
"""Provide L1 structural RTLIR generation pass."""
from __future__ import absolute_import, division, print_function

from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.rtlir.errors import RTLIRConversionError
from pymtl3.passes.rtlir.rtype.RTLIRType import get_rtlir

from .StructuralRTLIRSignalExpr import gen_signal_expr


class StructuralRTLIRGenL1Pass( BasePass ):
  def __init__( s, conns_self_self, conns_self_child, conns_child_child ):
    # connections_self_self, connections_self_child, connections_child_child
    s.c_ss = conns_self_self
    s.c_sc = conns_self_child
    s.c_cc = conns_child_child

  def __call__( s, tr_top ):
    """ generate structural RTLIR for component `tr_top` """
    if not hasattr( tr_top, '_pass_structural_rtlir_gen' ):
      tr_top._pass_structural_rtlir_gen = PassMetadata()
    s.tr_top = tr_top
    try:
      s.gen_rtlir_types( tr_top )
      s.gen_constants( tr_top )
      s.sort_connections( tr_top )
    except AssertionError as e:
      msg = '' if e.args[0] is None else e.args[0]
      raise RTLIRConversionError( tr_top, msg )

  def gen_rtlir_types( s, tr_top ):
    tr_top._pass_structural_rtlir_gen.rtlir_type = get_rtlir( tr_top )

  def gen_constants( s, m ):
    ns = m._pass_structural_rtlir_gen
    ns.consts = []
    rtype = ns.rtlir_type
    const_types = rtype.get_consts_packed()
    for const_name, const_rtype in const_types:
      assert hasattr(m, const_name), \
        "Internal error: {} is not a member of {}".format( const_name, m )
      const_instance = getattr(m, const_name)
      ns.consts.append( ( const_name, const_rtype, const_instance ) )

  def collect_connections( s, m ):
    return map( lambda x: \
      ((gen_signal_expr(m, x[0]), gen_signal_expr(m, x[1])), False), s.c_ss[m] )

  def sort_connections( s, m ):
    m_connections = s.collect_connections( m )
    connections = []
    for u, v in m.get_connect_order():
      _u, _v = gen_signal_expr( m, u ), gen_signal_expr( m, v )
      for idx, ( ( wr, rd ), visited ) in enumerate( m_connections ):

        if not visited and ( ( s.contains( _u, wr ) and s.contains( _v, rd ) ) or \
           ( s.contains( _u, rd ) and s.contains( _v, wr ) ) ):
          connections.append( ( wr, rd ) )
          m_connections[idx] = ( m_connections[idx][0], True )
    connections += map(lambda x:x[0],filter(lambda x:not x[1],m_connections))
    m._pass_structural_rtlir_gen.connections = connections

  def contains( s, obj, signal ):
    """Return if obj contains signal.

    At level 1 all signals have their corresponding object in `s.connect`.
    Therefore just checking whether obj is equal to signal is enough at level 1.
    """
    return obj == signal



#=========================================================================
# BehavioralRTLIRGenL4Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Oct 20, 2018
"""Provide L4 behavioral RTLIR generation pass."""
from __future__ import absolute_import, division, print_function

from pymtl3.passes.BasePass import BasePass, PassMetadata

from .BehavioralRTLIRGenL3Pass import BehavioralRTLIRGeneratorL3


class BehavioralRTLIRGenL4Pass( BasePass ):

  def __call__( s, m ):
    """ generate RTLIR for all upblks of m """
    if not hasattr( m, '_pass_behavioral_rtlir_gen' ):
      m._pass_behavioral_rtlir_gen = PassMetadata()
    m._pass_behavioral_rtlir_gen.rtlir_upblks = {}
    visitor = BehavioralRTLIRGeneratorL4( m )
    upblks = {
      'CombUpblk' : list(m.get_update_blocks() - m.get_update_on_edge()),
      'SeqUpblk'  : list(m.get_update_on_edge())
    }
    # Sort the upblks by their name
    upblks['CombUpblk'].sort( key = lambda x: x.__name__ )
    upblks['SeqUpblk'].sort( key = lambda x: x.__name__ )

    for upblk_type in ( 'CombUpblk', 'SeqUpblk' ):
      for blk in upblks[ upblk_type ]:
        visitor._upblk_type = upblk_type
        m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ] = \
          visitor.enter( blk, m.get_update_block_ast( blk ) )

class BehavioralRTLIRGeneratorL4( BehavioralRTLIRGeneratorL3 ):
  """Behavioral RTLIR generator level 4.

  Do nothing here because attributes have been handled in previous
  levels.
  """
  def __init__( s, component ):
    super( BehavioralRTLIRGeneratorL4, s ).__init__( component )



"""
========================================================================
SimpleSchedulePass.py
========================================================================
Generate a simple schedule (no Mamba techniques here) based on the
DAG generated by some previous pass.

Author : Shunning Jiang
Date   : Dec 26, 2018
"""
from __future__ import absolute_import, division, print_function

from collections import deque

from graphviz import Digraph

from pymtl3.dsl.errors import UpblkCyclicError

from .BasePass import BasePass, PassMetadata
from .errors import PassOrderError


class SimpleSchedulePass( BasePass ):
  def __call__( self, top ):
    if not hasattr( top._dag, "all_constraints" ):
      raise PassOrderError( "all_constraints" )

    top._sched = PassMetadata()

    top._sched.schedule = self.schedule( top )

  def schedule( self, top ):

    # Construct the graph

    V   = top._dag.final_upblks
    E   = top._dag.all_constraints
    Es  = { v: [] for v in V }
    InD = { v: 0  for v in V }

    for (u, v) in E: # u -> v
      InD[v] += 1
      Es [u].append( v )

    # Perform topological sort for a serial schedule.

    schedule = []

    Q = deque( [ v for v in V if not InD[v] ] )

    import random
    while Q:
      random.shuffle(Q)
      u = Q.pop()
      schedule.append( u )
      for v in Es[u]:
        InD[v] -= 1
        if not InD[v]:
          Q.append( v )

    import os

    check_schedule( top, schedule, V, E, InD )

    if 'MAMBA_DAG' in os.environ:
      dump_dag( top, V, E )

    return schedule

def dump_dag( top, V, E ):
  from graphviz import Digraph
  from pymtl3.dsl import CalleePort
  dot = Digraph()
  dot.graph_attr["rank"] = "same"
  dot.graph_attr["ratio"] = "compress"
  dot.graph_attr["margin"] = "0.1"

  for x in V:
    x_name = repr(x) if isinstance( x, CalleePort ) else x.__name__
    try:
      x_host = repr(x.get_parent_object() if isinstance( x, CalleePort )
                    else top.get_update_block_host_component(x))
    except:
      x_host = ""
    dot.node( x_name +"\\n@" + x_host, shape="box")

  for (x, y) in E:
    x_name = repr(x) if isinstance( x, CalleePort ) else x.__name__
    try:
      x_host = repr(x.get_parent_object() if isinstance( x, CalleePort )
                    else top.get_update_block_host_component(x))
    except:
      x_host = ""
    y_name = repr(y) if isinstance( y, CalleePort ) else y.__name__
    try:
      y_host = repr(y.get_parent_object() if isinstance( y, CalleePort )
                    else top.get_update_block_host_component(y))
    except:
      y_host = ""

    dot.edge( x_name+"\\n@"+x_host, y_name+"\\n@"+y_host )
  dot.render( "/tmp/upblk-dag.gv", view=True )

def check_schedule( top, schedule, V, E, in_degree ):

  if len(schedule) != len(V):
    V_leftovers = {  v for v in V if in_degree[v]  }
    E_leftovers = {  (x,y) for (x,y) in E
                         if x in V_leftovers and y in V_leftovers  }
    dump_dag( top, V_leftovers, E_leftovers )

    raise UpblkCyclicError( """
Update blocks have cyclic dependencies.
* Please consult update dependency graph for details."
    """)



#=========================================================================
# BehavioralRTLIRL2Pass_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Feb 2, 2019
"""Test the level 2 behavioral RTLIR passes.

The L2 generation, L2 type check, and visualization passes are invoked. The
generation pass results are verified against a reference AST.
"""

from __future__ import absolute_import, division, print_function

import pytest

from pymtl3.datatypes import Bits1, Bits2, Bits4, Bits32, BitStruct
from pymtl3.dsl import Component, InPort, OutPort, Wire
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRGenL2Pass import (
    BehavioralRTLIRGenL2Pass,
)
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRTypeCheckL2Pass import (
    BehavioralRTLIRTypeCheckL2Pass,
)
from pymtl3.passes.rtlir.errors import PyMTLSyntaxError, PyMTLTypeError
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.rtlir.util.test_utility import do_test, expected_failure


def local_do_test( m ):
  """Check if generated behavioral RTLIR is the same as reference."""
  m.elaborate()
  m.apply( BehavioralRTLIRGenL2Pass() )
  m.apply( BehavioralRTLIRTypeCheckL2Pass() )

  try:
    ref = m._rtlir_test_ref
    for blk in m.get_update_blocks():
      upblk = m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ]
      assert upblk == ref[ blk.__name__ ]
  except AttributeError:
    pass

#-------------------------------------------------------------------------
# PyMTL type errors
#-------------------------------------------------------------------------

def test_L2_add_component( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = Bits1( 1 ) + s
  with expected_failure( PyMTLTypeError, "rhs of binop should be signal/const" ):
    do_test( A() )

def test_L2_inv_component( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = ~s
  with expected_failure( PyMTLTypeError, "only applies to signals and consts" ):
    do_test( A() )

def test_L2_for_start_component( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        for i in range( s, 8, 1 ):
          s.out = ~Bits1( 1 )
  with expected_failure( PyMTLTypeError, "must be a constant expression" ):
    do_test( A() )

def test_L2_for_end_component( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        for i in range( 0, s, 1 ):
          s.out = ~Bits1( 1 )
  with expected_failure( PyMTLTypeError, "must be a constant expression" ):
    do_test( A() )

def test_L2_for_step_component( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        for i in range( 0, 8, s ):
          s.out = ~Bits1( 1 )
  with expected_failure( PyMTLTypeError, "must be a constant expression" ):
    do_test( A() )

def test_L2_if_cond_component( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        if s:
          s.out = Bits1( 1 )
        else:
          s.out = ~Bits1( 1 )
  with expected_failure( PyMTLTypeError, "condition of if must be a signal" ):
    do_test( A() )

def test_L2_ifexp_cond_component( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = Bits1(1) if s else ~Bits1(1)
  with expected_failure( PyMTLTypeError, "condition of if-exp must be a signal" ):
    do_test( A() )

def test_L2_ifexp_body_component( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = s if 1 else ~Bits1(1)
  with expected_failure( PyMTLTypeError, "body of if-exp must be a signal" ):
    do_test( A() )

def test_L2_ifexp_orelse_component( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = Bits1(1) if 1 else s
  with expected_failure( PyMTLTypeError, "else branch of if-exp must be a signal" ):
    do_test( A() )

def test_L2_if_cond_bool( do_test ):
  class B( BitStruct ):
    def __init__( s ):
      s.foo = Bits32( 42 )
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        if s.in_:
          s.out = Bits1(1)
        else:
          s.out = ~Bits1(1)
  with expected_failure( PyMTLTypeError, "cannot be converted to bool" ):
    do_test( A() )

def test_L2_for_step_zero( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        for i in range( 0, 4, 0 ):
          s.out = Bits1( 1 )
  with expected_failure( PyMTLTypeError, "step of for-loop cannot be zero" ):
    do_test( A() )

def test_L2_for_step_variable( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits2 )
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        for i in range( 0, 4, s.in_ ):
          s.out = Bits1( 1 )
  with expected_failure( PyMTLTypeError, "step of a for-loop must be a constant" ):
    do_test( A() )

def test_L2_ifexp_cond_bool( do_test ):
  class B( BitStruct ):
    def __init__( s ):
      s.foo = Bits32( 42 )
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = Bits1(1) if s.in_ else ~Bits1(1)
  with expected_failure( PyMTLTypeError, "cannot be converted to bool" ):
    do_test( A() )

def test_L2_ifexp_body_else_diff_type( do_test ):
  class B( BitStruct ):
    def __init__( s ):
      s.foo = Bits32( 42 )
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = Bits1(1) if 1 else s.in_
  with expected_failure( PyMTLTypeError, "must have the same type" ):
    do_test( A() )

def test_L2_unary_not_cast_to_bool( do_test ):
  class B( BitStruct ):
    def __init__( s ):
      s.foo = Bits32( 42 )
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = not s.in_
  with expected_failure( PyMTLTypeError, "cannot be cast to bool" ):
    do_test( A() )

def test_L2_bool_cast_to_bool( do_test ):
  class B( BitStruct ):
    def __init__( s ):
      s.foo = Bits32( 42 )
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = Bits1( 1 ) and s.in_
  with expected_failure( PyMTLTypeError, "cannot be cast into bool" ):
    do_test( A() )

def test_L2_binop_non_vector( do_test ):
  class B( BitStruct ):
    def __init__( s ):
      s.foo = Bits32( 42 )
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = Bits1( 1 ) + s.in_
  with expected_failure( PyMTLTypeError, "should be of vector type" ):
    do_test( A() )

#-------------------------------------------------------------------------
# PyMTL syntax errors
#-------------------------------------------------------------------------

def test_L2_call_bool( do_test ):
  Bool = rdt.Bool
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = Bool( Bits1(1) )
  with expected_failure( PyMTLSyntaxError, "bool values cannot be instantiated explicitly" ):
    do_test( A() )

def test_L2_tmp_var_used_before_assignement( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits4 )
      @s.update
      def upblk():
        s.out = u + Bits4( 1 )
  with expected_failure( PyMTLSyntaxError, "tmpvar u used before assignment" ):
    do_test( A() )

def test_L2_for_else( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits4 )
      @s.update
      def upblk():
        for i in range(4):
          s.out = Bits4( 1 )
        else:
          s.out = Bits4( 1 )
  with expected_failure( PyMTLSyntaxError, "for loops cannot have 'else' branch" ):
    do_test( A() )

def test_L2_for_index_signal( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = Wire( Bits4 )
      s.out = OutPort( Bits4 )
      @s.update
      def upblk():
        for s.in_ in range(4):
          s.out = Bits4( 1 )
  with expected_failure( PyMTLSyntaxError, "loop index must be a temporary variable" ):
    do_test( A() )

def test_L2_for_index_redefined( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits4 )
      @s.update
      def upblk():
        for i in range(4):
          for i in range(4):
            s.out = Bits4( 1 )
  with expected_failure( PyMTLSyntaxError, "Redefinition of loop index i" ):
    do_test( A() )

def test_L2_for_iter_name( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits2 )
      s.out = OutPort( Bits4 )
      @s.update
      def upblk():
        for i in s.in_:
          s.out = Bits4( 1 )
  with expected_failure( PyMTLSyntaxError, "only use (x)range() after 'in'" ):
    do_test( A() )

def test_L2_for_iter_call_non_range( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits4 )
      def foo(): pass
      @s.update
      def upblk():
        for i in foo():
          s.out = Bits4( 1 )
  with expected_failure( PyMTLSyntaxError, "only use (x)range() after 'in'" ):
    do_test( A() )

def test_L2_for_iter_range_arg_0( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits4 )
      @s.update
      def upblk():
        for i in range():
          s.out = Bits4( 1 )
  with expected_failure( PyMTLSyntaxError, "1~3 arguments should be given to (x)range" ):
    do_test( A() )

def test_L2_for_iter_range_arg_4( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits4 )
      @s.update
      def upblk():
        for i in range( 0, 4, 1, 1 ):
          s.out = Bits4( 1 )
  with expected_failure( PyMTLSyntaxError, "1~3 arguments should be given to (x)range" ):
    do_test( A() )

def test_L2_invalid_op_is( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits4 )
      @s.update
      def upblk():
        s.out = Bits1( 1 ) is Bits1( 1 )
  with expected_failure( PyMTLSyntaxError, "not supported" ):
    do_test( A() )

def test_L2_invalid_op_is_not( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits4 )
      @s.update
      def upblk():
        s.out = Bits1( 1 ) is not Bits1( 1 )
  with expected_failure( PyMTLSyntaxError, "not supported" ):
    do_test( A() )

def test_L2_invalid_op_in( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits4 )
      @s.update
      def upblk():
        s.out = Bits1( 1 ) in Bits1( 1 )
  with expected_failure( PyMTLSyntaxError, "not supported" ):
    do_test( A() )

def test_L2_invalid_op_not_in( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits4 )
      @s.update
      def upblk():
        s.out = Bits1( 1 ) not in Bits1( 1 )
  with expected_failure( PyMTLSyntaxError, "not supported" ):
    do_test( A() )

def test_L2_invalid_op_floor_div( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits4 )
      @s.update
      def upblk():
        s.out = Bits1( 1 ) // Bits1( 1 )
  with expected_failure( PyMTLSyntaxError, "not supported" ):
    do_test( A() )

def test_L2_compare_terms( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits4 )
      @s.update
      def upblk():
        s.out = Bits1( 0 ) <= Bits2( 1 ) <= Bits2( 2 )
  with expected_failure( PyMTLSyntaxError, "Comparison can only have 2 operands" ):
    do_test( A() )

#-------------------------------------------------------------------------
# PyMTL syntax errors for unsupported Python syntax
#-------------------------------------------------------------------------

def test_Break( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        for i in range(42): break
  with expected_failure( PyMTLSyntaxError, "invalid operation: break" ):
    do_test( A() )

def test_Continue( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        for i in range(42): continue
  with expected_failure( PyMTLSyntaxError, "invalid operation: continue" ):
    do_test( A() )



from __future__ import absolute_import, division, print_function

from copy import deepcopy

from pymtl3 import *


class RegisterFile( Component ):

  def construct( s, Type, nregs=32, rd_ports=1, wr_ports=1,
                const_zero=False ):

    addr_type = mk_bits( clog2( nregs ) )

    s.raddr = [ InPort( addr_type ) for i in xrange( rd_ports ) ]
    s.rdata = [ OutPort( Type ) for i in xrange( rd_ports ) ]

    s.waddr = [ InPort( addr_type ) for i in xrange( wr_ports ) ]
    s.wdata = [ InPort( Type ) for i in xrange( wr_ports ) ]
    s.wen   = [ InPort( Bits1 ) for i in xrange( wr_ports ) ]

    s.regs      = [ Wire( Type ) for i in xrange(nregs) ]
    s.next_regs = [ Wire( Type ) for i in xrange(nregs) ]

    @s.update_on_edge
    def up_rfile():
      for i in xrange( nregs ):
        s.regs[i] = s.next_regs[i]

    @s.update
    def up_rf_read():
      for i in xrange( rd_ports ):
        s.rdata[i] = s.regs[ s.raddr[i] ]

    if const_zero:
      @s.update
      def up_rf_write_constzero():
        for i in xrange( nregs ):
          s.next_regs[i] = deepcopy( s.regs[i] )
        for i in xrange( wr_ports ):
          if s.wen[i] & (s.waddr[i] != addr_type(0)):
            s.next_regs[ s.waddr[i] ] = deepcopy( s.wdata[i] )

    else:
      @s.update
      def up_rf_write():
        for i in xrange( nregs ):
          s.next_regs[i] = deepcopy( s.regs[i] )
        for i in xrange( wr_ports ):
          if s.wen[i]:
            s.next_regs[ s.waddr[i] ] = deepcopy( s.wdata[i] )



#=========================================================================
# SVBehavioralTranslatorL2_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 28, 2019
"""Test the SystemVerilog translator implementation."""

from __future__ import absolute_import, division, print_function

from pymtl3.datatypes import Bits1, Bits32, reduce_and, reduce_or, reduce_xor
from pymtl3.dsl import Component, InPort, OutPort
from pymtl3.passes.rtlir import BehavioralRTLIRGenPass, BehavioralRTLIRTypeCheckPass
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.passes.sverilog.translation.behavioral.SVBehavioralTranslatorL2 import (
    BehavioralRTLIRToSVVisitorL2,
)

from .SVBehavioralTranslatorL1_test import is_sverilog_reserved


def local_do_test( m ):
  m.elaborate()
  m.apply( BehavioralRTLIRGenPass() )
  m.apply( BehavioralRTLIRTypeCheckPass() )

  visitor = BehavioralRTLIRToSVVisitorL2(is_sverilog_reserved)
  upblks = m._pass_behavioral_rtlir_gen.rtlir_upblks
  m_all_upblks = m.get_update_blocks()
  for blk in m_all_upblks:
    upblk_src = visitor.enter( blk, upblks[blk] )
    upblk_src = "\n".join( upblk_src )
    assert upblk_src == m._ref_upblk_srcs[blk.__name__]

def test_reduce( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_1 = InPort( Bits32 )
      s.in_2 = InPort( Bits32 )
      s.in_3 = InPort( Bits32 )
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = reduce_and( s.in_1 ) & reduce_or( s.in_2 ) | reduce_xor( s.in_3 )
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = ( ( & in_1 ) & ( | in_2 ) ) | ( ^ in_3 );
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_1 = Bits32(tv[0])
    m.in_2 = Bits32(tv[1])
    m.in_3 = Bits32(tv[2])
  def tv_out( m, tv ):
    assert m.out == Bits1(tv[3])
  a._test_vectors = [
    [  0,   1,    2,  1   ],
    [ -1,   1,   -1,  1   ],
    [  9,   8,    7,  1   ],
    [  9,   8,    0,  0   ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_if( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_1 = InPort( Bits32 )
      s.in_2 = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        if Bits1(1):
          s.out = s.in_1
        else:
          s.out = s.in_2
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  if ( 1'd1 ) begin
    out = in_1;
  end
  else
    out = in_2;
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_1 = Bits32(tv[0])
    m.in_2 = Bits32(tv[1])
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[2])
  a._test_vectors = [
    [    0,    -1,   0 ],
    [   42,     0,  42 ],
    [   24,    42,  24 ],
    [   -2,    24,  -2 ],
    [   -1,    -2,  -1 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_if_dangling_else_inner( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_1 = InPort( Bits32 )
      s.in_2 = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        if Bits1(1):
          if Bits1(0):
            s.out = s.in_1
          else:
            s.out = s.in_2
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  if ( 1'd1 ) begin
    if ( 1'd0 ) begin
      out = in_1;
    end
    else
      out = in_2;
  end
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_1 = Bits32(tv[0])
    m.in_2 = Bits32(tv[1])
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[2])
  a._test_vectors = [
    [    0,    -1,  -1 ],
    [   42,     0,   0 ],
    [   24,    42,  42 ],
    [   -2,    24,  24 ],
    [   -1,    -2,  -2 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_if_dangling_else_outter( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_1 = InPort( Bits32 )
      s.in_2 = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        if Bits1(1):
          if Bits1(0):
            s.out = s.in_1
        else:
          s.out = s.in_2
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  if ( 1'd1 ) begin
    if ( 1'd0 ) begin
      out = in_1;
    end
  end
  else
    out = in_2;
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_1 = Bits32(tv[0])
    m.in_2 = Bits32(tv[1])
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[2])
  a._test_vectors = [
    [    0,    -1,   0 ],
    [   42,     0,   0 ],
    [   24,    42,   0 ],
    [   -2,    24,   0 ],
    [   -1,    -2,   0 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_if_branches( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_1 = InPort( Bits32 )
      s.in_2 = InPort( Bits32 )
      s.in_3 = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        if Bits1(1):
          s.out = s.in_1
        elif Bits1(0):
          s.out = s.in_2
        else:
          s.out = s.in_3
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  if ( 1'd1 ) begin
    out = in_1;
  end
  else if ( 1'd0 ) begin
    out = in_2;
  end
  else
    out = in_3;
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_1 = Bits32(tv[0])
    m.in_2 = Bits32(tv[1])
    m.in_3 = Bits32(tv[2])
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[3])
  a._test_vectors = [
    [    0,    -1,   0,  0 ],
    [   42,     0,  42, 42 ],
    [   24,    42,  24, 24 ],
    [   -2,    24,  -2, -2 ],
    [   -1,    -2,  -1, -1 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_nested_if( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_1 = InPort( Bits32 )
      s.in_2 = InPort( Bits32 )
      s.in_3 = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        if Bits1(1):
          if Bits1(0):
            s.out = s.in_1
          else:
            s.out = s.in_2
        elif Bits1(0):
          if Bits1(1):
            s.out = s.in_2
          else:
            s.out = s.in_3
        else:
          if Bits1(1):
            s.out = s.in_3
          else:
            s.out = s.in_1
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  if ( 1'd1 ) begin
    if ( 1'd0 ) begin
      out = in_1;
    end
    else
      out = in_2;
  end
  else if ( 1'd0 ) begin
    if ( 1'd1 ) begin
      out = in_2;
    end
    else
      out = in_3;
  end
  else if ( 1'd1 ) begin
    out = in_3;
  end
  else
    out = in_1;
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_1 = Bits32(tv[0])
    m.in_2 = Bits32(tv[1])
    m.in_3 = Bits32(tv[2])
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[3])
  a._test_vectors = [
    [    0,    -1,   0,  -1 ],
    [   42,     0,  42,   0 ],
    [   24,    42,  24,  42 ],
    [   -2,    24,  -2,  24 ],
    [   -1,    -2,  -1,  -2 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_for_range_upper( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range(2) ]
      s.out = [ OutPort( Bits32 ) for _ in range(2) ]
      @s.update
      def upblk():
        for i in range(2):
          s.out[i] = s.in_[i]
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  for ( int i = 0; i < 2; i += 1 )
    out[i] = in_[i];
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_[0] = Bits32(tv[0])
    m.in_[1] = Bits32(tv[1])
  def tv_out( m, tv ):
    assert m.out[0] == Bits32(tv[2])
    assert m.out[1] == Bits32(tv[3])
  a._test_vectors = [
    [    0,    -1,   0,  -1 ],
    [   42,     0,  42,   0 ],
    [   24,    42,  24,  42 ],
    [   -2,    24,  -2,  24 ],
    [   -1,    -2,  -1,  -2 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { 'upblk' : \
"""\
integer __loopvar_upblk$i;

always_comb begin : upblk
  for ( __loopvar_upblk$i = 0; __loopvar_upblk$i < 2; __loopvar_upblk$i = __loopvar_upblk$i + 1 )
    out[__loopvar_upblk$i] = in_[__loopvar_upblk$i];
end\
""" }
  do_test( a )

def test_for_range_lower_upper( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range(2) ]
      s.out = [ OutPort( Bits32 ) for _ in range(2) ]
      @s.update
      def upblk():
        for i in range(1, 2):
          s.out[i] = s.in_[i]
        s.out[0] = s.in_[0]
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  for ( int i = 1; i < 2; i += 1 )
    out[i] = in_[i];
  out[0] = in_[0];
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_[0] = Bits32(tv[0])
    m.in_[1] = Bits32(tv[1])
  def tv_out( m, tv ):
    assert m.out[0] == Bits32(tv[2])
    assert m.out[1] == Bits32(tv[3])
  a._test_vectors = [
    [    0,    -1,   0,  -1 ],
    [   42,     0,  42,   0 ],
    [   24,    42,  24,  42 ],
    [   -2,    24,  -2,  24 ],
    [   -1,    -2,  -1,  -2 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { 'upblk' : \
"""\
integer __loopvar_upblk$i;

always_comb begin : upblk
  for ( __loopvar_upblk$i = 1; __loopvar_upblk$i < 2; __loopvar_upblk$i = __loopvar_upblk$i + 1 )
    out[__loopvar_upblk$i] = in_[__loopvar_upblk$i];
  out[0] = in_[0];
end\
""" }
  do_test( a )

def test_for_range_lower_upper_step( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range(5) ]
      s.out = [ OutPort( Bits32 ) for _ in range(5) ]
      @s.update
      def upblk():
        for i in range(0, 5, 2):
          s.out[i] = s.in_[i]
        for i in range(1, 5, 2):
          s.out[i] = s.in_[i]
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  for ( int i = 0; i < 5; i += 2 )
    out[i] = in_[i];
  for ( int i = 1; i < 5; i += 2 )
    out[i] = in_[i];
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_[0] = Bits32(tv[0])
    m.in_[1] = Bits32(tv[1])
    m.in_[2] = Bits32(tv[2])
    m.in_[3] = Bits32(tv[3])
    m.in_[4] = Bits32(tv[4])
  def tv_out( m, tv ):
    assert m.out[0] == Bits32(tv[5])
    assert m.out[1] == Bits32(tv[6])
    assert m.out[2] == Bits32(tv[7])
    assert m.out[3] == Bits32(tv[8])
    assert m.out[4] == Bits32(tv[9])
  a._test_vectors = [
    [    0,    -1,   0,  -1,   0,    0,    -1,   0,  -1,   0, ],
    [   42,     0,  42,   0,  42,   42,     0,  42,   0,  42, ],
    [   24,    42,  24,  42,  24,   24,    42,  24,  42,  24, ],
    [   -2,    24,  -2,  24,  -2,   -2,    24,  -2,  24,  -2, ],
    [   -1,    -2,  -1,  -2,  -1,   -1,    -2,  -1,  -2,  -1, ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { 'upblk' : \
"""\
integer __loopvar_upblk$i;

always_comb begin : upblk
  for ( __loopvar_upblk$i = 0; __loopvar_upblk$i < 5; __loopvar_upblk$i = __loopvar_upblk$i + 2 )
    out[__loopvar_upblk$i] = in_[__loopvar_upblk$i];
  for ( __loopvar_upblk$i = 1; __loopvar_upblk$i < 5; __loopvar_upblk$i = __loopvar_upblk$i + 2 )
    out[__loopvar_upblk$i] = in_[__loopvar_upblk$i];
end\
""" }
  do_test( a )

def test_if_exp_for( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range(5) ]
      s.out = [ OutPort( Bits32 ) for _ in range(5) ]
      @s.update
      def upblk():
        for i in range(5):
          s.out[i] = s.in_[i] if i == 1 else s.in_[0]
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  for ( int i = 0; i < 5; i += 1 )
    out[i] = ( i == 1 ) ? in_[i] : in_[0];
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_[0] = Bits32(tv[0])
    m.in_[1] = Bits32(tv[1])
    m.in_[2] = Bits32(tv[2])
    m.in_[3] = Bits32(tv[3])
    m.in_[4] = Bits32(tv[4])
  def tv_out( m, tv ):
    assert m.out[0] == Bits32(tv[5])
    assert m.out[1] == Bits32(tv[6])
    assert m.out[2] == Bits32(tv[7])
    assert m.out[3] == Bits32(tv[8])
    assert m.out[4] == Bits32(tv[9])
  a._test_vectors = [
    [    0,    -1,   0,  -1,   0,    0,    -1,   0,   0,   0, ],
    [   42,     0,  42,   0,  42,   42,     0,  42,  42,  42, ],
    [   24,    42,  24,  42,  24,   24,    42,  24,  24,  24, ],
    [   -2,    24,  -2,  24,  -2,   -2,    24,  -2,  -2,  -2, ],
    [   -1,    -2,  -1,  -2,  -1,   -1,    -2,  -1,  -1,  -1, ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { 'upblk' : \
"""\
integer __loopvar_upblk$i;

always_comb begin : upblk
  for ( __loopvar_upblk$i = 0; __loopvar_upblk$i < 5; __loopvar_upblk$i = __loopvar_upblk$i + 1 )
    out[__loopvar_upblk$i] = ( __loopvar_upblk$i == 1 ) ? in_[__loopvar_upblk$i] : in_[0];
end\
""" }
  do_test( a )

def test_if_exp_unary_op( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range(5) ]
      s.out = [ OutPort( Bits32 ) for _ in range(5) ]
      @s.update
      def upblk():
        for i in range(5):
          s.out[i] = (~s.in_[i]) if i == 1 else s.in_[0]
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  for ( int i = 0; i < 5; i += 1 )
    out[i] = ( i == 1 ) ? ~in_[i] : in_[0];
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_[0] = Bits32(tv[0])
    m.in_[1] = Bits32(tv[1])
    m.in_[2] = Bits32(tv[2])
    m.in_[3] = Bits32(tv[3])
    m.in_[4] = Bits32(tv[4])
  def tv_out( m, tv ):
    assert m.out[0] == Bits32(tv[5])
    assert m.out[1] == Bits32(tv[6])
    assert m.out[2] == Bits32(tv[7])
    assert m.out[3] == Bits32(tv[8])
    assert m.out[4] == Bits32(tv[9])
  a._test_vectors = [
    [    0,    -1,   0,  -1,   0,    0,    ~-1,   0,   0,   0, ],
    [   42,     0,  42,   0,  42,   42,     ~0,  42,  42,  42, ],
    [   24,    42,  24,  42,  24,   24,    ~42,  24,  24,  24, ],
    [   -2,    24,  -2,  24,  -2,   -2,    ~24,  -2,  -2,  -2, ],
    [   -1,    -2,  -1,  -2,  -1,   -1,    ~-2,  -1,  -1,  -1, ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { 'upblk' : \
"""\
integer __loopvar_upblk$i;

always_comb begin : upblk
  for ( __loopvar_upblk$i = 0; __loopvar_upblk$i < 5; __loopvar_upblk$i = __loopvar_upblk$i + 1 )
    out[__loopvar_upblk$i] = ( __loopvar_upblk$i == 1 ) ? ~in_[__loopvar_upblk$i] : in_[0];
end\
""" }
  do_test( a )

def test_if_bool_op( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range(5) ]
      s.out = [ OutPort( Bits32 ) for _ in range(5) ]
      @s.update
      def upblk():
        for i in range(5):
          if ( s.in_[i] != Bits32(0) ) and ( ( s.in_[i+1] != Bits32(0) ) if i<4 else ( s.in_[4] != Bits32(0) )):
            s.out[i] = s.in_[i]
          else:
            s.out[i] = Bits32(0)
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  for ( int i = 0; i < 5; i += 1 )
    if ( ( in_[i] != 32'd0 ) && ( ( i < 4 ) ? in_[i + 1] != 32'd0 : in_[4] != 32'd0 ) ) begin
      out[i] = in_[i];
    end
    else
      out[i] = 32'd0;
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_[0] = Bits32(tv[0])
    m.in_[1] = Bits32(tv[1])
    m.in_[2] = Bits32(tv[2])
    m.in_[3] = Bits32(tv[3])
    m.in_[4] = Bits32(tv[4])
  def tv_out( m, tv ):
    assert m.out[0] == Bits32(tv[5])
    assert m.out[1] == Bits32(tv[6])
    assert m.out[2] == Bits32(tv[7])
    assert m.out[3] == Bits32(tv[8])
    assert m.out[4] == Bits32(tv[9])
  a._test_vectors = [
    [    0,    -1,   0,  -1,   0,     0,     0,    0,    0,    0, ],
    [   42,     0,  42,   0,  42,     0,     0,    0,    0,   42, ],
    [   24,    42,  24,  42,  24,    24,    42,   24,   42,   24, ],
    [   -2,    24,  -2,  24,  -2,    -2,    24,   -2,   24,   -2, ],
    [   -1,    -2,  -1,  -2,  -1,    -1,    -2,   -1,   -2,   -1, ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { 'upblk' : \
"""\
integer __loopvar_upblk$i;

always_comb begin : upblk
  for ( __loopvar_upblk$i = 0; __loopvar_upblk$i < 5; __loopvar_upblk$i = __loopvar_upblk$i + 1 )
    if ( ( in_[__loopvar_upblk$i] != 32'd0 ) && ( ( __loopvar_upblk$i < 4 ) ? in_[__loopvar_upblk$i + 1] != 32'd0 : in_[4] != 32'd0 ) ) begin
      out[__loopvar_upblk$i] = in_[__loopvar_upblk$i];
    end
    else
      out[__loopvar_upblk$i] = 32'd0;
end\
""" }
  do_test( a )

def test_tmpvar( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range(5) ]
      s.out = [ OutPort( Bits32 ) for _ in range(5) ]
      @s.update
      def upblk():
        for i in range(5):
          if ( s.in_[i]  != Bits32(0) ) and ( ( s.in_[i+1] != Bits32(0) ) if i<4 else ( s.in_[4] != Bits32(0) ) ):
            tmpvar = s.in_[i]
          else:
            tmpvar = Bits32(0)
          s.out[i] = tmpvar
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  for ( int i = 0; i < 5; i += 1 ) begin
    if ( ( in_[i] != 32'd0 ) && ( ( i < 4 ) ? in_[i + 1] != 32'd0 : in_[4] != 32'd0 ) ) begin
      __tmpvar_upblk$tmpvar = in_[i];
    end
    else
      __tmpvar_upblk$tmpvar = 32'd0;
    out[i] = __tmpvar_upblk$tmpvar;
  end
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_[0] = Bits32(tv[0])
    m.in_[1] = Bits32(tv[1])
    m.in_[2] = Bits32(tv[2])
    m.in_[3] = Bits32(tv[3])
    m.in_[4] = Bits32(tv[4])
  def tv_out( m, tv ):
    assert m.out[0] == Bits32(tv[5])
    assert m.out[1] == Bits32(tv[6])
    assert m.out[2] == Bits32(tv[7])
    assert m.out[3] == Bits32(tv[8])
    assert m.out[4] == Bits32(tv[9])
  a._test_vectors = [
    [    0,    -1,   0,  -1,   0,     0,     0,    0,    0,    0, ],
    [   42,     0,  42,   0,  42,     0,     0,    0,    0,   42, ],
    [   24,    42,  24,  42,  24,    24,    42,   24,   42,   24, ],
    [   -2,    24,  -2,  24,  -2,    -2,    24,   -2,   24,   -2, ],
    [   -1,    -2,  -1,  -2,  -1,    -1,    -2,   -1,   -2,   -1, ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { 'upblk' : \
"""\
integer __loopvar_upblk$i;

always_comb begin : upblk
  for ( __loopvar_upblk$i = 0; __loopvar_upblk$i < 5; __loopvar_upblk$i = __loopvar_upblk$i + 1 ) begin
    if ( ( in_[__loopvar_upblk$i] != 32'd0 ) && ( ( __loopvar_upblk$i < 4 ) ? in_[__loopvar_upblk$i + 1] != 32'd0 : in_[4] != 32'd0 ) ) begin
      __tmpvar_upblk$tmpvar = in_[__loopvar_upblk$i];
    end
    else
      __tmpvar_upblk$tmpvar = 32'd0;
    out[__loopvar_upblk$i] = __tmpvar_upblk$tmpvar;
  end
end\
""" }
  do_test( a )



"""
========================================================================
helpers.py
========================================================================
Useful helper functions of Bits operations. Ideally these functions
should be implemented in RPython.

Author : Shunning Jiang
Date   : Nov 3, 2017
"""
from __future__ import absolute_import, division, print_function

import math

from .bits_import import *


def concat( *args ):
  nbits = sum( [ x.nbits for x in args ] )
  concat_bits = Bits( nbits, 0 )

  begin = 0
  for bits in reversed( args ):
    concat_bits[ begin : begin+bits.nbits ] = bits
    begin += bits.nbits

  return concat_bits

def zext( value, new_width ):
  assert new_width > value.nbits
  return Bits( new_width, value )

def clog2( N ):
  assert N > 0
  return int( math.ceil( math.log( N, 2 ) ) )

def sext( value, new_width ):
  assert new_width > value.nbits
  return Bits( new_width, value.int() )

def reduce_and( value ):
  try:
    return b1( int(value) == (1 << value.nbits) - 1 )
  except AttributeError:
    raise TypeError("Cannot call reduce_and on int")

def reduce_or( value ):
  try:
    return b1( int(value) != 0 )
  except AttributeError:
    raise TypeError("Cannot call reduce_or on int")

def reduce_xor( value ):
  try:
    pop_count = 0
    value = int(value)
    while value != 0:
      pop_count += value & 1
      value >>= 1
    return b1( pop_count & 1 )

  except AttributeError:
    raise TypeError("Cannot call reduce_xor on int")



#=========================================================================
# SVBehavioralTranslatorL3_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 28, 2019
"""Test the SystemVerilog translator implementation."""

from __future__ import absolute_import, division, print_function

from pymtl3.datatypes import Bits32, Bits96, BitStruct, concat
from pymtl3.dsl import Component, InPort, OutPort
from pymtl3.passes.rtlir import BehavioralRTLIRGenPass, BehavioralRTLIRTypeCheckPass
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.passes.sverilog.translation.behavioral.SVBehavioralTranslatorL3 import (
    BehavioralRTLIRToSVVisitorL3,
)

from .SVBehavioralTranslatorL1_test import is_sverilog_reserved


def local_do_test( m ):
  m.elaborate()
  m.apply( BehavioralRTLIRGenPass() )
  m.apply( BehavioralRTLIRTypeCheckPass() )

  visitor = BehavioralRTLIRToSVVisitorL3(is_sverilog_reserved)
  upblks = m._pass_behavioral_rtlir_gen.rtlir_upblks
  m_all_upblks = m.get_update_blocks()
  for blk in m_all_upblks:
    upblk_src = visitor.enter( blk, upblks[blk] )
    upblk_src = "\n".join( upblk_src )
    assert upblk_src == m._ref_upblk_srcs[blk.__name__]

def test_struct( do_test ):
  class B( BitStruct ):
    def __init__( s, foo=42 ):
      s.foo = Bits32(foo)
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.in_.foo
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = in_.foo;
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = tv[0]
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[1])
  a._test_vectors = [
    [     B(),  42 ],
    [    B(0),   0 ],
    [   B(-1),  -1 ],
    [   B(42),  42 ],
    [   B(-2),  -2 ],
    [   B(10),  10 ],
    [  B(256), 256 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { 'upblk' : \
"""\
always_comb begin : upblk
  out = in_$foo;
end\
""" }
  do_test( a )

def test_struct_const( do_test ):
  class B( BitStruct ):
    fields = [ ( 'foo', Bits32 ) ]
    def __init__( s, foo=42 ):
      s.foo = Bits32(foo)
  class A( Component ):
    def construct( s ):
      s.in_ = B()
      s.out = OutPort( Bits32 )
      s.out_b = OutPort( B )
      @s.update
      def upblk():
        s.out = s.in_.foo
        s.out_b = s.in_
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = in_.foo;
  out_b = in_;
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    pass
  def tv_out( m, tv ):
    assert m.out_b == tv[0]
    assert m.out == Bits32(tv[1])
  a._test_vectors = [
    [       B(),   42 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { 'upblk' : \
"""\
always_comb begin : upblk
  out = 32'd42;
  out_b = 32'd42;
end\
""" }
  do_test( a )

def test_packed_array_behavioral( do_test ):
  class B( BitStruct ):
    def __init__( s, foo=42, bar=1 ):
      s.foo = Bits32(foo)
      s.bar = [ Bits32(bar) for _ in range(2) ]
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits96 )
      @s.update
      def upblk():
        s.out = concat( s.in_.bar[0], s.in_.bar[1], s.in_.foo )
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = { in_.bar[0], in_.bar[1], in_.foo };
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = tv[0]
  def tv_out( m, tv ):
    assert m.out == Bits96(tv[1])
  a._test_vectors = [
    [        B(),   concat(   Bits32(1),   Bits32(1),  Bits32(42) ) ],
    [    B(0, 0),   concat(   Bits32(0),   Bits32(0),   Bits32(0) ) ],
    [  B(-1, -1),   concat(  Bits32(-1),  Bits32(-1),  Bits32(-1) ) ],
    [  B(-1, 42),   concat(  Bits32(42),  Bits32(42),  Bits32(-1) ) ],
    [  B(42, 42),   concat(  Bits32(42),  Bits32(42),  Bits32(42) ) ],
    [  B(42, -1),   concat(  Bits32(-1),  Bits32(-1),  Bits32(42) ) ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { 'upblk' : \
"""\
always_comb begin : upblk
  out = { in_$bar[0], in_$bar[1], in_$foo };
end\
""" }
  do_test( a )

def test_nested_struct( do_test ):
  class C( BitStruct ):
    def __init__( s, woof=2 ):
      s.woof = Bits32(woof)
  class B( BitStruct ):
    def __init__( s, foo=42, bar=1 ):
      s.foo = Bits32(foo)
      s.bar = [ Bits32(bar) for _ in range(2) ]
      s.c = C()
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits96 )
      @s.update
      def upblk():
        s.out = concat( s.in_.bar[0], s.in_.c.woof, s.in_.foo )
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = { in_.bar[0], in_.c.woof, in_.foo };
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = tv[0]
  def tv_out( m, tv ):
    assert m.out == Bits96(tv[1])
  a._test_vectors = [
    [        B(),   concat(   Bits32(1),   Bits32(2),  Bits32(42) ) ],
    [    B(0, 0),   concat(   Bits32(0),   Bits32(2),   Bits32(0) ) ],
    [  B(-1, -1),   concat(  Bits32(-1),   Bits32(2),  Bits32(-1) ) ],
    [  B(-1, 42),   concat(  Bits32(42),   Bits32(2),  Bits32(-1) ) ],
    [  B(42, 42),   concat(  Bits32(42),   Bits32(2),  Bits32(42) ) ],
    [  B(42, -1),   concat(  Bits32(-1),   Bits32(2),  Bits32(42) ) ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { 'upblk' : \
"""\
always_comb begin : upblk
  out = { in_$bar[0], in_$c$woof, in_$foo };
end\
""" }
  do_test( a )



#=========================================================================
# test_utility.py
#=========================================================================
# Author : Peitian Pan
# Date   : Feb 21, 2019
"""Test utilities used by RTLIR tests."""

from __future__ import absolute_import, division, print_function

from contextlib import contextmanager

import pytest


@pytest.fixture
def do_test( request ):
  """Call `local_do_test` of the requesting module."""
  return request.module.local_do_test

@contextmanager
def expected_failure( exception = Exception, msg = None ):
  """Mark one test case as should-fail.

  Not to be confused with pytest.xfail, which is commonly used to mark
  tests related to unimplemented functionality. This test only passes when
  it throws an expected exception.
  """
  try:
    yield
  except exception as e:
    if msg is None or e.args[0].find( msg ) != -1:
      return
    else:
      raise
  raise Exception( 'expected-to-fail test unexpectedly passed!' )



#=========================================================================
# StructuralTranslatorL1_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 21, 2019
"""Test the level 1 structural translators."""

from __future__ import absolute_import, division, print_function

from pymtl3.datatypes import Bits1, Bits4, Bits16, Bits32
from pymtl3.dsl import Component, InPort, OutPort, Wire
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir.errors import RTLIRConversionError
from pymtl3.passes.rtlir.util.test_utility import do_test, expected_failure
from pymtl3.passes.translator.structural.StructuralTranslatorL1 import (
    StructuralTranslatorL1,
)

from .TestStructuralTranslator import mk_TestStructuralTranslator


def local_do_test( m ):
  m.elaborate()
  tr = mk_TestStructuralTranslator(StructuralTranslatorL1)(m)
  tr.clear( m )
  tr.translate_structural(m)
  try:
    name = tr.structural.component_unique_name[m]
    assert name == m._ref_name
    decl_ports = tr.structural.decl_ports[m]
    assert decl_ports == m._ref_ports
    decl_wires = tr.structural.decl_wires[m]
    assert decl_wires == m._ref_wires
    decl_consts = tr.structural.decl_consts[m]
    assert decl_consts == m._ref_consts
    connections = tr.structural.connections[m]
    assert connections == m._ref_conns
    vector_types = tr.structural.decl_type_vector
    assert sorted(vector_types, key=lambda x: str(x[0])) == m._ref_vectors
  except AttributeError:
    pass

def test_component_args( do_test ):
  class A( Component ):
    def construct( s, foo, bar ):
      assert foo == Bits4(0)
      assert bar == Bits16(42)
  a = A( Bits4(0), Bits16(42) )
  a._ref_name = 'A__foo_0__bar_002a'
  a._ref_src = \
"""\
component {}
(
port_decls:
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:

endcomponent
""".format( a._ref_name )
  do_test( a )

def test_component_dfefault_args( do_test ):
  class A( Component ):
    def construct( s, foo, bar=Bits16(42) ):
      assert foo == Bits4(0)
      assert bar == Bits16(42)
  a = A( Bits4(0) )
  a._ref_name = 'A__foo_0__bar_002a'
  a._ref_src = \
"""\
component {}
(
port_decls:
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:

endcomponent
""".format( a._ref_name )
  do_test( a )

def test_component_kw_args( do_test ):
  class A( Component ):
    def construct( s, foo, bar ):
      assert foo == Bits4(0)
      assert bar == Bits16(42)
  a = A( foo = Bits4(0), bar = Bits16(42) )
  a._ref_name = 'A__foo_0__bar_002a'
  a._ref_src = \
"""\
component {}
(
port_decls:
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:

endcomponent
""".format( a._ref_name )
  do_test( a )

def test_component_star_args( do_test ):
  class A( Component ):
    def construct( s, *args ):
      assert args[0] == Bits4(0)
      assert args[1] == Bits16(42)
  args = [ Bits4(0), Bits16(42) ]
  a = A( *args )
  with expected_failure( RTLIRConversionError, "varargs are not allowed" ):
    do_test( a )

def test_component_star_args_ungroup( do_test ):
  class A( Component ):
    def construct( s, foo, bar ):
      assert foo == Bits4(0)
      assert bar == Bits16(42)
  args = [ Bits4(0), Bits16(42) ]
  a = A( *args )
  a._ref_name = 'A__foo_0__bar_002a'
  a._ref_src = \
"""\
component {}
(
port_decls:
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:

endcomponent
""".format( a._ref_name )
  do_test( a )

def test_component_double_star_args( do_test ):
  class A( Component ):
    def construct( s, **kwargs ):
      assert kwargs['foo'] == Bits4(0)
      assert kwargs['bar'] == Bits16(42)
  kwargs = { 'foo':Bits4(0), 'bar':Bits16(42) }
  a = A( **kwargs )
  with expected_failure( RTLIRConversionError, "keyword args are not allowed" ):
    do_test( a )

def test_component_double_star_args_ungroup( do_test ):
  class A( Component ):
    def construct( s, foo, bar ):
      assert foo == Bits4(0)
      assert bar == Bits16(42)
  kwargs = { 'foo':Bits4(0), 'bar':Bits16(42) }
  a = A( **kwargs )
  a._ref_name = 'A__foo_0__bar_002a'
  a._ref_src = \
"""\
component {}
(
port_decls:
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:

endcomponent
""".format( a._ref_name )
  do_test( a )

def test_component_mixed_kw_args( do_test ):
  class A( Component ):
    def construct( s, foo, bar, woo=Bits32(0) ):
      pass
  a = A( Bits4(0), bar = Bits16(42) )
  a._ref_name = 'A__foo_0__bar_002a__woo_00000000'
  a._ref_src = \
"""\
component {}
(
port_decls:
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:

endcomponent
""".format( a._ref_name )
  do_test( a )

def test_port_decls( do_test ):
  class A( Component ):
    def construct( s ):
      s.foo = InPort( Bits32 )
      s.bar = InPort( Bits4 )
  a = A()
  a._ref_name = "A"
  a._ref_ports = \
"""\
port_decls:
  port_decl: bar Port of Vector4
  port_decl: foo Port of Vector32
"""
  a._ref_src = \
"""\
component {}
(
port_decls:
  port_decl: bar Port of Vector4
  port_decl: foo Port of Vector32
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:

endcomponent
""".format( a._ref_name )
  do_test( a )

def test_array_port_decl( do_test ):
  class A( Component ):
    def construct( s ):
      s.foo = [ InPort( Bits32 ) for _ in range(5) ]
  a = A()
  a._ref_name = "A"
  a._ref_ports = \
"""\
port_decls:
  port_decl: foo Array[5] of Port
"""
  a._ref_src = \
"""\
component {}
(
port_decls:
  port_decl: foo Array[5] of Port
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:

endcomponent
""".format( a._ref_name )
  do_test( a )

def test_wire_decls( do_test ):
  class A( Component ):
    def construct( s ):
      s.foo = Wire( Bits32 )
      s.bar = Wire( Bits4 )
      @s.update
      def upblk():
        s.foo = 42
        s.bar = Bits4(0)
  a = A()
  a._ref_name = "A"
  a._ref_ports = "port_decls:\n"
  a._ref_wires = \
"""\
wire_decls:
  wire_decl: bar Wire of Vector4
  wire_decl: foo Wire of Vector32
"""
  a._ref_src = \
"""\
component {}
(
port_decls:
interface_decls:
);
const_decls:
freevars:
wire_decls:
  wire_decl: bar Wire of Vector4
  wire_decl: foo Wire of Vector32
component_decls:
tmpvars:
upblk_srcs:
  upblk_src: upblk
connections:

endcomponent
""".format( a._ref_name )
  do_test( a )

def test_array_wire_decl( do_test ):
  class A( Component ):
    def construct( s ):
      s.foo = [ Wire( Bits32 ) for _ in range(5) ]
      @s.update
      def upblk():
        for i in range(5):
          s.foo[i] = Bits32(0)
  a = A()
  a._ref_name = "A"
  a._ref_ports = "port_decls:\n"
  a._ref_wires = \
"""\
wire_decls:
  wire_decl: foo Array[5] of Wire
"""
  a._ref_src = \
"""\
component {}
(
port_decls:
interface_decls:
);
const_decls:
freevars:
wire_decls:
  wire_decl: foo Array[5] of Wire
component_decls:
tmpvars:
upblk_srcs:
  upblk_src: upblk
connections:

endcomponent
""".format( a._ref_name )
  do_test( a )

def test_const_decls( do_test ):
  class A( Component ):
    def construct( s ):
      s.foo = Bits32(0)
      s.bar = Bits4(0)
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.foo
        s.out = Bits32(s.bar)
  a = A()
  a._ref_name = "A"
  a._ref_ports = \
"""\
port_decls:
  port_decl: out Port of Vector32
"""
  a._ref_wires = "wire_decls:\n"
  a._ref_consts = \
"""\
const_decls:
  const_decl: bar Const of Vector4
  const_decl: foo Const of Vector32
"""
  a._ref_src = \
"""\
component {}
(
port_decls:
  port_decl: out Port of Vector32
interface_decls:
);
const_decls:
  const_decl: bar Const of Vector4
  const_decl: foo Const of Vector32
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
  upblk_src: upblk
connections:

endcomponent
""".format( a._ref_name )
  do_test( a )

def test_array_const_decl( do_test ):
  class A( Component ):
    def construct( s ):
      s.foo = [ Bits32(0) for _ in range(5) ]
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.foo[0]
  a = A()
  a._ref_name = "A"
  a._ref_ports = \
"""\
port_decls:
  port_decl: out Port of Vector32
"""
  a._ref_wires = "wire_decls:\n"
  a._ref_consts = \
"""\
const_decls:
  const_decl: foo Array[5] of Const
"""
  a._ref_src = \
"""\
component {}
(
port_decls:
  port_decl: out Port of Vector32
interface_decls:
);
const_decls:
  const_decl: foo Array[5] of Const
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
  upblk_src: upblk
connections:

endcomponent
""".format( a._ref_name )
  do_test( a )

def test_expr_bit_sel( do_test ):
  # PyMTL DSL treat bit selection as a 1-bit part selection!
  class A( Component ):
    def construct( s ):
      s.foo = InPort( Bits32 )
      s.bar = OutPort( Bits1 )
      s.connect( s.bar, s.foo[1] )
  a = A()
  a._ref_name = "A"
  a._ref_ports = \
"""\
port_decls:
  port_decl: bar Port of Vector1
  port_decl: foo Port of Vector32
"""
  a._ref_wires = "wire_decls:\n"
  a._ref_consts = "const_decls:\n"
  a._ref_conns = \
"""\
connections:
  connection: PartSel CurCompAttr foo 1 2 -> CurCompAttr bar
"""
  a._ref_src = \
"""\
component {}
(
port_decls:
  port_decl: bar Port of Vector1
  port_decl: foo Port of Vector32
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:
  connection: PartSel CurCompAttr foo 1 2 -> CurCompAttr bar

endcomponent
""".format( a._ref_name )
  do_test( a )

def test_expr_part_sel( do_test ):
  class A( Component ):
    def construct( s ):
      s.foo = InPort( Bits32 )
      s.bar = OutPort( Bits4 )
      s.connect( s.bar, s.foo[0:4] )
  a = A()
  a._ref_name = "A"
  a._ref_ports = \
"""\
port_decls:
  port_decl: bar Port of Vector4
  port_decl: foo Port of Vector32
"""
  a._ref_wires = "wire_decls:\n"
  a._ref_consts = "const_decls:\n"
  a._ref_conns = \
"""\
connections:
  connection: PartSel CurCompAttr foo 0 4 -> CurCompAttr bar
"""
  a._ref_src = \
"""\
component {}
(
port_decls:
  port_decl: bar Port of Vector4
  port_decl: foo Port of Vector32
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:
  connection: PartSel CurCompAttr foo 0 4 -> CurCompAttr bar

endcomponent
""".format( a._ref_name )
  do_test( a )

def test_expr_port_array_index( do_test ):
  class A( Component ):
    def construct( s ):
      s.foo = [ InPort( Bits32 ) for _ in range(5) ]
      s.bar = OutPort( Bits32 )
      s.connect( s.bar, s.foo[1] )
  a = A()
  a._ref_name = "A"
  a._ref_ports = \
"""\
port_decls:
  port_decl: bar Port of Vector32
  port_decl: foo Array[5] of Port
"""
  a._ref_wires = "wire_decls:\n"
  a._ref_consts = "const_decls:\n"
  a._ref_conns = \
"""\
connections:
  connection: PortArrayIdx CurCompAttr foo 1 -> CurCompAttr bar
"""
  a._ref_src = \
"""\
component {}
(
port_decls:
  port_decl: bar Port of Vector32
  port_decl: foo Array[5] of Port
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:
  connection: PortArrayIdx CurCompAttr foo 1 -> CurCompAttr bar

endcomponent
""".format( a._ref_name )
  do_test( a )

def test_expr_wire_array_index( do_test ):
  class A( Component ):
    def construct( s ):
      s.wire = [ Wire( Bits32 ) for _ in range(5) ]
      s.bar = OutPort( Bits32 )
      s.connect( s.bar, s.wire[1] )
      @s.update
      def upblk():
        for i in range(5):
          s.wire[i] = Bits32(0)
  a = A()
  a._ref_name = "A"
  a._ref_ports = \
"""\
port_decls:
  port_decl: bar Port of Vector32
"""
  a._ref_wires = \
"""\
wire_decls:
  wire_decl: wire Array[5] of Wire
"""
  a._ref_consts = "const_decls:\n"
  a._ref_conns = \
"""\
connections:
  connection: WireArrayIdx CurCompAttr wire 1 -> CurCompAttr bar
"""
  a._ref_src = \
"""\
component {}
(
port_decls:
  port_decl: bar Port of Vector32
interface_decls:
);
const_decls:
freevars:
wire_decls:
  wire_decl: wire Array[5] of Wire
component_decls:
tmpvars:
upblk_srcs:
  upblk_src: upblk
connections:
  connection: WireArrayIdx CurCompAttr wire 1 -> CurCompAttr bar

endcomponent
""".format( a._ref_name )
  do_test( a )

def test_expr_const_array_index( do_test ):
  # Since PyMTL DSL does not provide names for constants, the
  # strutural RTLIR will remove the signal hierarchy for constants
  # and just treat them as constant integers.
  class A( Component ):
    def construct( s ):
      s.const = [ 0 for _ in range(5) ]
      s.bar = OutPort( Bits32 )
      s.connect( s.bar, s.const[1] )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.const[0]
  a = A()
  a._ref_name = "A"
  a._ref_ports = \
"""\
port_decls:
  port_decl: bar Port of Vector32
  port_decl: out Port of Vector32
"""
  a._ref_wires = "wire_decls:\n"
  a._ref_consts = \
"""\
const_decls:
  const_decl: const Array[5] of Const
"""
  a._ref_conns = \
"""\
connections:
  connection: Bits32(0) -> CurCompAttr bar
"""
  a._ref_src = \
"""\
component {}
(
port_decls:
  port_decl: bar Port of Vector32
  port_decl: out Port of Vector32
interface_decls:
);
const_decls:
  const_decl: const Array[5] of Const
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
  upblk_src: upblk
connections:
  connection: Bits32(0) -> CurCompAttr bar

endcomponent
""".format( a._ref_name )
  do_test( a )

def test_dtype_vector( do_test ):
  class A( Component ):
    def construct( s ):
      s.foo = InPort(Bits32)
  a = A()
  a._ref_name = "A"
  a._ref_ports = \
"""\
port_decls:
  port_decl: foo Port of Vector32
"""
  a._ref_wires = "wire_decls:\n"
  a._ref_consts = "const_decls:\n"
  a._ref_conns = "connections:\n"
  a._ref_vectors = [ (rdt.Vector(1), "Vector1"), (rdt.Vector(32), "Vector32") ]
  a._ref_src = \
"""\
component {}
(
port_decls:
  port_decl: foo Port of Vector32
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:

endcomponent
""".format( a._ref_name )
  do_test( a )

__all__ = list(filter(lambda s: s.startswith('test_'), dir()))



#=========================================================================
# BehavioralRTLIRGenL3Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Oct 20, 2018
"""Provide L3 behavioral RTLIR generation pass."""
from __future__ import absolute_import, division, print_function

from pymtl3.datatypes import Bits, BitStruct
from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.rtlir.errors import PyMTLSyntaxError

from . import BehavioralRTLIR as bir
from .BehavioralRTLIRGenL2Pass import BehavioralRTLIRGeneratorL2


class BehavioralRTLIRGenL3Pass( BasePass ):
  def __call__( s, m ):
    """ generate RTLIR for all upblks of m """
    if not hasattr( m, '_pass_behavioral_rtlir_gen' ):
      m._pass_behavioral_rtlir_gen = PassMetadata()
    m._pass_behavioral_rtlir_gen.rtlir_upblks = {}
    visitor = BehavioralRTLIRGeneratorL3( m )

    upblks = {
      'CombUpblk' : list(m.get_update_blocks() - m.get_update_on_edge()),
      'SeqUpblk'  : list(m.get_update_on_edge())
    }
    # Sort the upblks by their name
    upblks['CombUpblk'].sort( key = lambda x: x.__name__ )
    upblks['SeqUpblk'].sort( key = lambda x: x.__name__ )

    for upblk_type in ( 'CombUpblk', 'SeqUpblk' ):
      for blk in upblks[ upblk_type ]:
        visitor._upblk_type = upblk_type
        m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ] = \
          visitor.enter( blk, m.get_update_block_ast( blk ) )

class BehavioralRTLIRGeneratorL3( BehavioralRTLIRGeneratorL2 ):
  def __init__( s, component ):
    super( BehavioralRTLIRGeneratorL3, s ).__init__( component )

  def visit_Call( s, node ):
    """Return behavioral RTLIR of a method call.

    At L3 we need to support the syntax of struct instantiation in upblks.
    This is achieved by function calls like `struct( 1, 2, 0 )`.
    """
    obj = s.get_call_obj( node )
    if isinstance( obj, type ) and issubclass( obj, BitStruct ):
      if len(node.args) < 1:
        raise PyMTLSyntaxError( s.blk, node,
          'at least one value should be provided to struct instantiation!' )
      values = map(lambda arg: s.visit(arg), node.args)
      ret = bir.StructInst( obj, values )
      ret.ast = node
      return ret

    else:
      return super( BehavioralRTLIRGeneratorL3, s ).visit_Call( node )
